//
// edip_param_c.hpp
//
// LGPL Version 2.1 HEADER START
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
//
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
// MA 02110-1301  USA
//
// LGPL Version 2.1 HEADER END
//

//
// Copyright (c) 2021, Regents of the University of Minnesota.
// All rights reserved.
//
// Contributors:
//    Yaser Afshar
//

#ifndef EDIP_PARAM_C_HPP
#define EDIP_PARAM_C_HPP

/*!
 * \brief EDIPC parameter class
 *
 * Reference:
 *  Marks, N. A., "Generalizing the environment-dependent interaction potential
 *  for carbon," Phys. Rev.B, 63(3):035401, Dec 2000.
 */
struct EDIPCParam {
  /*! Two-body parameter in eV (EQ. 2) */
  double eps;
  /*! Two-body parameter B^4 (where B = 0.9538 angstroms) (EQ. 2) */
  double BB;
  /*! Two-body parameter beta (EQ. 2) */
  double beta;
  /*! Two-body parameter sigma in angstroms (EQ. 2) */
  double sigma;
  /*! Two-body parameter a in angstroms (EQ. 2) */
  double a;
  /*! Two-body parameter aprime in angstroms (EQ. 2) */
  double aprime;

  /*! Three-body parameter Z0 (EQs. 3 & 4) */
  double Z0;
  /*! Three-body parameter lambda0 in eV (EQs. 3 & 4)  */
  double lambda0;
  /*! Three-body parameter lambdaprime (EQs. 3 & 4) */
  double lambdaprime;
  /*! Three-body parameter gamma in angstroms (EQs. 3 & 5) */
  double gamma;
  /*! Three-body parameter q (EQs. 3 & 6) */
  double q;

  // Coordination parameters
  /*! The three-parameter function lower limit (in angstroms) (EQ. 8) */
  double flow;
  /*! The three-parameter function upper limit (in angstroms) (EQ. 8) */
  double fhigh;
  /*! The three-parameter function parameter alpha (EQ. 8) */
  double alpha;
  /*!
   * \brief The three-parameter function lower limit (in angstroms) (EQ. 8)
   * \note Paper says 1.481
   */
  double plow;
  /*! The three-parameter function upper limit (in angstroms) (EQ. 8) */
  double phigh;
  /*!
   * \brief Coordination counting parameter (in eV) (EQ. 10)
   *
   * It is fitted to Hartree-Fock cluster data, with the remainder determined
   * from density-functional theory (DFT) calculations of the
   * graphite-diamond transformation.
   */
  double Zdih;
  /*! Coordination parameter in eV (EQ. 11 & 12) */
  double Zrep;
  /*! Repulsive cutoff limit in angstroms (EQ. 14) */
  double c0;

  /*! Cutoff skin */
  double skin{0.2};
};

#endif  // EDIP_PARAM_C
